import { BLANKLINE_END_REGEX, BLANKLINE_START_REGEX, LINEBREAK_REGEX, NON_ALPHA_NUMERIC_REGEX, WHITESPACE_REGEX } from "../constants";
import { DiffOperation } from "../types";
import { math } from "../utils";
import { PatchObject } from "./PatchObject";
export class DiffMatchPatch {
    diffTimeout = 1.0;
    diffEditCost = 4;
    matchThreshold = 0.5;
    matchDistance = 1000;
    patchDeleteThreshold = 0.5;
    patchMargin = 4;
    matchMaxBits = 32;
    diff_main(text1, text2, optChecklines, optDeadline) {
        if (typeof optDeadline === "undefined") {
            if (this.diffTimeout <= 0) {
                optDeadline = Number.MAX_VALUE;
            }
            else {
                optDeadline = Date.now() + this.diffTimeout * 1000;
            }
        }
        const deadline = optDeadline;
        if (text1 == null || text2 == null) {
            throw new Error("Null input. (diff_main)");
        }
        if (text1 === text2) {
            if (text1) {
                return [[DiffOperation.DIFF_EQUAL, text1]];
            }
            return [];
        }
        if (typeof optChecklines === "undefined") {
            optChecklines = true;
        }
        const checklines = optChecklines;
        let commonlength = this.diff_commonPrefix(text1, text2);
        const commonprefix = text1.substring(0, commonlength);
        text1 = text1.substring(commonlength);
        text2 = text2.substring(commonlength);
        commonlength = this.diff_commonSuffix(text1, text2);
        const commonsuffix = text1.substring(text1.length - commonlength);
        text1 = text1.substring(0, text1.length - commonlength);
        text2 = text2.substring(0, text2.length - commonlength);
        const diffs = this.diff_compute_(text1, text2, checklines, deadline);
        if (commonprefix) {
            diffs.unshift([DiffOperation.DIFF_EQUAL, commonprefix]);
        }
        if (commonsuffix) {
            diffs.push([DiffOperation.DIFF_EQUAL, commonsuffix]);
        }
        this.diff_cleanupMerge(diffs);
        return diffs;
    }
    diff_commonPrefix(text1, text2) {
        if (!text1 ||
            !text2 ||
            text1.charAt(0) !== text2.charAt(0)) {
            return 0;
        }
        let pointermin = 0;
        let pointermax = math.min(text1.length, text2.length);
        let pointermid = pointermax;
        let pointerstart = 0;
        while (pointermin < pointermid) {
            if (text1.substring(pointerstart, pointermid) ===
                text2.substring(pointerstart, pointermid)) {
                pointermin = pointermid;
                pointerstart = pointermin;
            }
            else {
                pointermax = pointermid;
            }
            pointermid = Math.floor((pointermax - pointermin) / 2 + pointermin);
        }
        return pointermid;
    }
    diff_commonSuffix(text1, text2) {
        if (!text1 ||
            !text2 ||
            text1.charAt(text1.length - 1) !== text2.charAt(text2.length - 1)) {
            return 0;
        }
        let pointermin = 0;
        let pointermax = math.min(text1.length, text2.length);
        let pointermid = pointermax;
        let pointerend = 0;
        while (pointermin < pointermid) {
            if (text1.substring(text1.length - pointermid, text1.length - pointerend) ===
                text2.substring(text2.length - pointermid, text2.length - pointerend)) {
                pointermin = pointermid;
                pointerend = pointermin;
            }
            else {
                pointermax = pointermid;
            }
            pointermid = Math.floor((pointermax - pointermin) / 2 + pointermin);
        }
        return pointermid;
    }
    diff_cleanupSemantic(diffs) {
        let changes = false;
        const equalities = [];
        let equalitiesLength = 0;
        let lastEquality = null;
        let pointer = 0;
        let lengthInsertions1 = 0;
        let lengthDeletions1 = 0;
        let lengthInsertions2 = 0;
        let lengthDeletions2 = 0;
        while (pointer < diffs.length) {
            if (diffs[pointer][0] === DiffOperation.DIFF_EQUAL) {
                equalities[equalitiesLength++] = pointer;
                lengthInsertions1 = lengthInsertions2;
                lengthDeletions1 = lengthDeletions2;
                lengthInsertions2 = 0;
                lengthDeletions2 = 0;
                lastEquality = diffs[pointer][1];
            }
            else {
                if (diffs[pointer][0] === DiffOperation.DIFF_INSERT) {
                    lengthInsertions2 += diffs[pointer][1].length;
                }
                else {
                    lengthDeletions2 += diffs[pointer][1].length;
                }
                if (lastEquality &&
                    (lastEquality.length <= math.max(lengthInsertions1, lengthDeletions1)) &&
                    (lastEquality.length <= math.max(lengthInsertions2, lengthDeletions2))) {
                    diffs.splice(equalities[equalitiesLength - 1], 0, [DiffOperation.DIFF_DELETE, lastEquality]);
                    diffs[equalities[equalitiesLength - 1] + 1][0] = DiffOperation.DIFF_INSERT;
                    equalitiesLength--;
                    equalitiesLength--;
                    pointer = equalitiesLength > 0 ? equalities[equalitiesLength - 1] : -1;
                    lengthInsertions1 = 0;
                    lengthDeletions1 = 0;
                    lengthInsertions2 = 0;
                    lengthDeletions2 = 0;
                    lastEquality = null;
                    changes = true;
                }
            }
            pointer++;
        }
        if (changes) {
            this.diff_cleanupMerge(diffs);
        }
        this.diff_cleanupSemanticLossless(diffs);
        pointer = 1;
        while (pointer < diffs.length) {
            if (diffs[pointer - 1][0] === DiffOperation.DIFF_DELETE &&
                diffs[pointer][0] === DiffOperation.DIFF_INSERT) {
                const deletion = diffs[pointer - 1][1];
                const insertion = diffs[pointer][1];
                const overlapLength1 = this.diff_commonOverlap_(deletion, insertion);
                const overlapLength2 = this.diff_commonOverlap_(insertion, deletion);
                if (overlapLength1 >= overlapLength2) {
                    if (overlapLength1 >= deletion.length / 2 ||
                        overlapLength1 >= insertion.length / 2) {
                        diffs.splice(pointer, 0, [DiffOperation.DIFF_EQUAL, insertion.substring(0, overlapLength1)]);
                        diffs[pointer - 1][1] = deletion.substring(0, deletion.length - overlapLength1);
                        diffs[pointer + 1][1] = insertion.substring(overlapLength1);
                        pointer++;
                    }
                }
                else {
                    if (overlapLength2 >= deletion.length / 2 ||
                        overlapLength2 >= insertion.length / 2) {
                        diffs.splice(pointer, 0, [DiffOperation.DIFF_EQUAL, deletion.substring(0, overlapLength2)]);
                        diffs[pointer - 1][0] = DiffOperation.DIFF_INSERT;
                        diffs[pointer - 1][1] = insertion.substring(0, insertion.length - overlapLength2);
                        diffs[pointer + 1][0] = DiffOperation.DIFF_DELETE;
                        diffs[pointer + 1][1] = deletion.substring(overlapLength2);
                        pointer++;
                    }
                }
                pointer++;
            }
            pointer++;
        }
    }
    diff_cleanupSemanticLossless(diffs) {
        let pointer = 1;
        while (pointer < diffs.length - 1) {
            if (diffs[pointer - 1][0] === DiffOperation.DIFF_EQUAL &&
                diffs[pointer + 1][0] === DiffOperation.DIFF_EQUAL) {
                let equality1 = diffs[pointer - 1][1];
                let edit = diffs[pointer][1];
                let equality2 = diffs[pointer + 1][1];
                const commonOffset = this.diff_commonSuffix(equality1, edit);
                if (commonOffset) {
                    const commonString = edit.substring(edit.length - commonOffset);
                    equality1 = equality1.substring(0, equality1.length - commonOffset);
                    edit = commonString + edit.substring(0, edit.length - commonOffset);
                    equality2 = commonString + equality2;
                }
                let bestEquality1 = equality1;
                let bestEdit = edit;
                let bestEquality2 = equality2;
                let bestScore = this.diff_cleanupSemanticScore_(equality1, edit)
                    + this.diff_cleanupSemanticScore_(edit, equality2);
                while (edit.charAt(0) === equality2.charAt(0)) {
                    equality1 += edit.charAt(0);
                    edit = edit.substring(1) + equality2.charAt(0);
                    equality2 = equality2.substring(1);
                    const score = this.diff_cleanupSemanticScore_(equality1, edit)
                        + this.diff_cleanupSemanticScore_(edit, equality2);
                    if (score >= bestScore) {
                        bestScore = score;
                        bestEquality1 = equality1;
                        bestEdit = edit;
                        bestEquality2 = equality2;
                    }
                }
                if (diffs[pointer - 1][1] !== bestEquality1) {
                    if (bestEquality1) {
                        diffs[pointer - 1][1] = bestEquality1;
                    }
                    else {
                        diffs.splice(pointer - 1, 1);
                        pointer--;
                    }
                    diffs[pointer][1] = bestEdit;
                    if (bestEquality2) {
                        diffs[pointer + 1][1] = bestEquality2;
                    }
                    else {
                        diffs.splice(pointer + 1, 1);
                        pointer--;
                    }
                }
            }
            pointer++;
        }
    }
    diff_cleanupEfficiency(diffs) {
        let changes = false;
        const equalities = [];
        let equalitiesLength = 0;
        let lastEquality = null;
        let pointer = 0;
        let preIns = false;
        let preDel = false;
        let postIns = false;
        let postDel = false;
        while (pointer < diffs.length) {
            if (diffs[pointer][0] === DiffOperation.DIFF_EQUAL) {
                if (diffs[pointer][1].length < this.diffEditCost &&
                    (postIns || postDel)) {
                    equalities[equalitiesLength++] = pointer;
                    preIns = postIns;
                    preDel = postDel;
                    lastEquality = diffs[pointer][1];
                }
                else {
                    equalitiesLength = 0;
                    lastEquality = null;
                }
                postIns = postDel = false;
            }
            else {
                if (diffs[pointer][0] === DiffOperation.DIFF_DELETE) {
                    postDel = true;
                }
                else {
                    postIns = true;
                }
                if (lastEquality &&
                    ((preIns && preDel && postIns && postDel) ||
                        ((lastEquality.length < this.diffEditCost / 2) &&
                            (Number(preIns) + Number(preDel) + Number(postIns) + Number(postDel)) === 3))) {
                    diffs.splice(equalities[equalitiesLength - 1], 0, [DiffOperation.DIFF_DELETE, lastEquality]);
                    diffs[equalities[equalitiesLength - 1] + 1][0] = DiffOperation.DIFF_INSERT;
                    equalitiesLength--;
                    lastEquality = null;
                    if (preIns && preDel) {
                        postIns = postDel = true;
                        equalitiesLength = 0;
                    }
                    else {
                        equalitiesLength--;
                        pointer = equalitiesLength > 0 ? equalities[equalitiesLength - 1] : -1;
                        postIns = postDel = false;
                    }
                    changes = true;
                }
            }
            pointer++;
        }
        if (changes) {
            this.diff_cleanupMerge(diffs);
        }
    }
    diff_cleanupMerge(diffs) {
        diffs.push([DiffOperation.DIFF_EQUAL, ""]);
        let pointer = 0;
        let countDelete = 0;
        let countInsert = 0;
        let textDelete = "";
        let textInsert = "";
        let commonlength;
        while (pointer < diffs.length) {
            switch (diffs[pointer][0]) {
                case DiffOperation.DIFF_INSERT:
                    countInsert++;
                    textInsert += diffs[pointer][1];
                    pointer++;
                    break;
                case DiffOperation.DIFF_DELETE:
                    countDelete++;
                    textDelete += diffs[pointer][1];
                    pointer++;
                    break;
                case DiffOperation.DIFF_EQUAL:
                    if (countDelete + countInsert > 1) {
                        if (countDelete !== 0 && countInsert !== 0) {
                            commonlength = this.diff_commonPrefix(textInsert, textDelete);
                            if (commonlength !== 0) {
                                if ((pointer - countDelete - countInsert) > 0 &&
                                    (diffs[pointer - countDelete - countInsert - 1][0]
                                        === DiffOperation.DIFF_EQUAL)) {
                                    diffs[pointer - countDelete - countInsert - 1][1]
                                        += textInsert.substring(0, commonlength);
                                }
                                else {
                                    diffs.splice(0, 0, [DiffOperation.DIFF_EQUAL, textInsert.substring(0, commonlength)]);
                                    pointer++;
                                }
                                textInsert = textInsert.substring(commonlength);
                                textDelete = textDelete.substring(commonlength);
                            }
                            commonlength = this.diff_commonSuffix(textInsert, textDelete);
                            if (commonlength !== 0) {
                                diffs[pointer][1] = textInsert.substring(textInsert.length
                                    - commonlength) + diffs[pointer][1];
                                textInsert = textInsert.substring(0, textInsert.length - commonlength);
                                textDelete = textDelete.substring(0, textDelete.length - commonlength);
                            }
                        }
                        pointer -= countDelete + countInsert;
                        diffs.splice(pointer, countDelete + countInsert);
                        if (textDelete.length) {
                            diffs.splice(pointer, 0, [DiffOperation.DIFF_DELETE, textDelete]);
                            pointer++;
                        }
                        if (textInsert.length) {
                            diffs.splice(pointer, 0, [DiffOperation.DIFF_INSERT, textInsert]);
                            pointer++;
                        }
                        pointer++;
                    }
                    else if (pointer !== 0 && diffs[pointer - 1][0] === DiffOperation.DIFF_EQUAL) {
                        diffs[pointer - 1][1] += diffs[pointer][1];
                        diffs.splice(pointer, 1);
                    }
                    else {
                        pointer++;
                    }
                    countInsert = 0;
                    countDelete = 0;
                    textDelete = "";
                    textInsert = "";
                    break;
            }
        }
        if (diffs[diffs.length - 1][1] === "") {
            diffs.pop();
        }
        let changes = false;
        pointer = 1;
        while (pointer < diffs.length - 1) {
            if (diffs[pointer - 1][0] === DiffOperation.DIFF_EQUAL &&
                diffs[pointer + 1][0] === DiffOperation.DIFF_EQUAL) {
                if (diffs[pointer][1].substring(diffs[pointer][1].length - diffs[pointer - 1][1].length)
                    === diffs[pointer - 1][1]) {
                    diffs[pointer][1] = diffs[pointer - 1][1]
                        + diffs[pointer][1].substring(0, diffs[pointer][1].length - diffs[pointer - 1][1].length);
                    diffs[pointer + 1][1] = diffs[pointer - 1][1] + diffs[pointer + 1][1];
                    diffs.splice(pointer - 1, 1);
                    changes = true;
                }
                else if (diffs[pointer][1].substring(0, diffs[pointer + 1][1].length)
                    === diffs[pointer + 1][1]) {
                    diffs[pointer - 1][1] += diffs[pointer + 1][1];
                    diffs[pointer][1] = diffs[pointer][1].substring(diffs[pointer + 1][1].length)
                        + diffs[pointer + 1][1];
                    diffs.splice(pointer + 1, 1);
                    changes = true;
                }
            }
            pointer++;
        }
        if (changes) {
            this.diff_cleanupMerge(diffs);
        }
    }
    diff_xIndex(diffs, loc) {
        let chars1 = 0;
        let chars2 = 0;
        let lastChars1 = 0;
        let lastChars2 = 0;
        let x;
        for (x = 0; x < diffs.length; x++) {
            if (diffs[x][0] !== DiffOperation.DIFF_INSERT) {
                chars1 += diffs[x][1].length;
            }
            if (diffs[x][0] !== DiffOperation.DIFF_DELETE) {
                chars2 += diffs[x][1].length;
            }
            if (chars1 > loc) {
                break;
            }
            lastChars1 = chars1;
            lastChars2 = chars2;
        }
        if (diffs.length !== x &&
            diffs[x][0] === DiffOperation.DIFF_DELETE) {
            return lastChars2;
        }
        return lastChars2 + (loc - lastChars1);
    }
    diff_prettyHtml(diffs) {
        const html = [];
        const patternAMP = /&/g;
        const patternLT = /</g;
        const patternGT = />/g;
        const patternPARA = /\n/g;
        for (let x = 0; x < diffs.length; x++) {
            const op = diffs[x][0];
            const data = diffs[x][1];
            const text = data.replace(patternAMP, "&amp;")
                .replace(patternLT, "&lt;")
                .replace(patternGT, "&gt;")
                .replace(patternPARA, "&para;<br>");
            switch (op) {
                case DiffOperation.DIFF_INSERT:
                    html[x] = '<ins style="background:#e6ffe6;">' + text + "</ins>";
                    break;
                case DiffOperation.DIFF_DELETE:
                    html[x] = '<del style="background:#ffe6e6;">' + text + "</del>";
                    break;
                case DiffOperation.DIFF_EQUAL:
                    html[x] = "<span>" + text + "</span>";
                    break;
            }
        }
        return html.join("");
    }
    diff_text1(diffs) {
        const text = [];
        for (let x = 0; x < diffs.length; x++) {
            if (diffs[x][0] !== DiffOperation.DIFF_INSERT) {
                text[x] = diffs[x][1];
            }
        }
        return text.join("");
    }
    diff_text2(diffs) {
        const text = [];
        for (let x = 0; x < diffs.length; x++) {
            if (diffs[x][0] !== DiffOperation.DIFF_DELETE) {
                text[x] = diffs[x][1];
            }
        }
        return text.join("");
    }
    diff_levenshtein(diffs) {
        let levenshtein = 0;
        let insertions = 0;
        let deletions = 0;
        for (let x = 0; x < diffs.length; x++) {
            const op = diffs[x][0];
            const data = diffs[x][1];
            switch (op) {
                case DiffOperation.DIFF_INSERT:
                    insertions += data.length;
                    break;
                case DiffOperation.DIFF_DELETE:
                    deletions += data.length;
                    break;
                case DiffOperation.DIFF_EQUAL:
                    levenshtein += math.max(insertions, deletions);
                    insertions = 0;
                    deletions = 0;
                    break;
            }
        }
        levenshtein += math.max(insertions, deletions);
        return levenshtein;
    }
    diff_toDelta(diffs) {
        const text = [];
        for (let x = 0; x < diffs.length; x++) {
            switch (diffs[x][0]) {
                case DiffOperation.DIFF_INSERT:
                    text[x] = "+" + encodeURI(diffs[x][1]);
                    break;
                case DiffOperation.DIFF_DELETE:
                    text[x] = "-" + diffs[x][1].length;
                    break;
                case DiffOperation.DIFF_EQUAL:
                    text[x] = "=" + diffs[x][1].length;
                    break;
            }
        }
        return text.join("\t").replace(/%20/g, " ");
    }
    diff_fromDelta(text1, delta) {
        const diffs = [];
        let diffsLength = 0;
        let pointer = 0;
        const tokens = delta.split(/\t/g);
        for (let x = 0; x < tokens.length; x++) {
            const param = tokens[x].substring(1);
            switch (tokens[x].charAt(0)) {
                case "+":
                    try {
                        diffs[diffsLength++] = [DiffOperation.DIFF_INSERT, decodeURI(param)];
                    }
                    catch {
                        throw new Error("Illegal escape in diff_fromDelta: " + param);
                    }
                    break;
                case "-":
                case "=":
                    const n = parseInt(param, 10);
                    if (isNaN(n) || n < 0) {
                        throw new Error("Invalid number in diff_fromDelta: " + param);
                    }
                    const text = text1.substring(pointer, pointer += n);
                    if (tokens[x].charAt(0) === "=") {
                        diffs[diffsLength++] = [DiffOperation.DIFF_EQUAL, text];
                    }
                    else {
                        diffs[diffsLength++] = [DiffOperation.DIFF_DELETE, text];
                    }
                    break;
                default:
                    if (tokens[x]) {
                        throw new Error("Invalid diff operation in diff_fromDelta: " + tokens[x]);
                    }
            }
        }
        if (pointer !== text1.length) {
            throw new Error("Delta length (" + pointer + ") does not equal source text length ("
                + text1.length + ")");
        }
        return diffs;
    }
    diff_linesToChars(text1, text2) {
        const lineArray = [];
        const lineHash = {};
        lineArray[0] = "";
        const chars1 = this.diff_linesToCharsMunge_(text1, lineArray, lineHash, 40000);
        const chars2 = this.diff_linesToCharsMunge_(text2, lineArray, lineHash, 65535);
        return { chars1, chars2, lineArray };
    }
    diff_charsToLines(diffs, lineArray) {
        for (let i = 0; i < diffs.length; i++) {
            const chars = diffs[i][1];
            const text = [];
            for (let j = 0; j < chars.length; j++) {
                text[j] = lineArray[chars.charCodeAt(j)];
            }
            diffs[i][1] = text.join("");
        }
    }
    match_main(text, pattern, loc) {
        if (text == null || pattern == null || loc == null) {
            throw new Error("Null input. (match_main)");
        }
        loc = math.max(0, math.min(loc, text.length));
        if (text === pattern) {
            return 0;
        }
        else if (!text.length) {
            return -1;
        }
        else if (text.substring(loc, loc + pattern.length) === pattern) {
            return loc;
        }
        else {
            return this.match_bitap_(text, pattern, loc);
        }
    }
    patch_make(a, optB, optC) {
        let text1;
        let diffs;
        if (typeof a === "string" &&
            typeof optB === "string" &&
            typeof optC === "undefined") {
            text1 = a;
            diffs = this.diff_main(text1, optB, true);
            if (diffs.length > 2) {
                this.diff_cleanupSemantic(diffs);
                this.diff_cleanupEfficiency(diffs);
            }
        }
        else if (a &&
            typeof a === "object" &&
            typeof optB === "undefined" &&
            typeof optC === "undefined") {
            diffs = a;
            text1 = this.diff_text1(diffs);
        }
        else if (typeof a === "string" &&
            optB &&
            typeof optB === "object" &&
            typeof optC === "undefined") {
            text1 = a;
            diffs = optB;
        }
        else if (typeof a === "string" &&
            typeof optB === "string" &&
            optC &&
            typeof optC === "object") {
            text1 = a;
            diffs = optC;
        }
        else {
            throw new Error("Unknown call format to patch_make");
        }
        if (diffs.length === 0) {
            return [];
        }
        const patches = [];
        let patch = new PatchObject();
        let patchDiffLength = 0;
        let charCount1 = 0;
        let charCount2 = 0;
        let prepatchText = text1;
        let postpatchText = text1;
        for (let x = 0; x < diffs.length; x++) {
            const diffType = diffs[x][0];
            const diffText = diffs[x][1];
            if (!patchDiffLength && diffType !== DiffOperation.DIFF_EQUAL) {
                patch.start1 = charCount1;
                patch.start2 = charCount2;
            }
            switch (diffType) {
                case DiffOperation.DIFF_INSERT:
                    patch.diffs[patchDiffLength++] = diffs[x];
                    patch.length2 += diffText.length;
                    postpatchText = postpatchText.substring(0, charCount2)
                        + diffText + postpatchText.substring(charCount2);
                    break;
                case DiffOperation.DIFF_DELETE:
                    patch.length1 += diffText.length;
                    patch.diffs[patchDiffLength++] = diffs[x];
                    postpatchText = postpatchText.substring(0, charCount2)
                        + postpatchText.substring(charCount2 + diffText.length);
                    break;
                case DiffOperation.DIFF_EQUAL:
                    if (diffText.length <= 2 * this.patchMargin &&
                        patchDiffLength &&
                        diffs.length !== x + 1) {
                        patch.diffs[patchDiffLength++] = diffs[x];
                        patch.length1 += diffText.length;
                        patch.length2 += diffText.length;
                    }
                    else if (diffText.length >= 2 * this.patchMargin) {
                        if (patchDiffLength) {
                            this.patch_addContext_(patch, prepatchText);
                            patches.push(patch);
                            patch = new PatchObject();
                            patchDiffLength = 0;
                            prepatchText = postpatchText;
                            charCount1 = charCount2;
                        }
                    }
                    break;
            }
            if (diffType !== DiffOperation.DIFF_INSERT) {
                charCount1 += diffText.length;
            }
            if (diffType !== DiffOperation.DIFF_DELETE) {
                charCount2 += diffText.length;
            }
        }
        if (patchDiffLength) {
            this.patch_addContext_(patch, prepatchText);
            patches.push(patch);
        }
        return patches;
    }
    patch_deepCopy(patches) {
        const patchesCopy = [];
        for (let x = 0; x < patches.length; x++) {
            const patch = patches[x];
            const patchCopy = new PatchObject();
            for (let y = 0; y < patch.diffs.length; y++) {
                patchCopy.diffs[y] = [patch.diffs[y][0], patch.diffs[y][1]];
            }
            patchCopy.start1 = patch.start1;
            patchCopy.start2 = patch.start2;
            patchCopy.length1 = patch.length1;
            patchCopy.length2 = patch.length2;
            patchesCopy[x] = patchCopy;
        }
        return patchesCopy;
    }
    patch_apply(patches, text) {
        if (patches.length === 0) {
            return [text, []];
        }
        patches = this.patch_deepCopy(patches);
        const nullPadding = this.patch_addPadding(patches);
        text = nullPadding + text + nullPadding;
        this.patch_splitMax(patches);
        let delta = 0;
        const results = [];
        for (let x = 0; x < patches.length; x++) {
            const expectedLoc = patches[x].start2 + delta;
            const text1 = this.diff_text1(patches[x].diffs);
            let startLoc;
            let endLoc = -1;
            if (text1.length > this.matchMaxBits) {
                startLoc = this.match_main(text, text1.substring(0, this.matchMaxBits), expectedLoc);
                if (startLoc !== -1) {
                    endLoc = this.match_main(text, text1.substring(text1.length - this.matchMaxBits), expectedLoc + text1.length - this.matchMaxBits);
                    if (endLoc === -1 || startLoc >= endLoc) {
                        startLoc = -1;
                    }
                }
            }
            else {
                startLoc = this.match_main(text, text1, expectedLoc);
            }
            if (startLoc === -1) {
                results[x] = false;
                delta -= patches[x].length2 - patches[x].length1;
            }
            else {
                results[x] = true;
                delta = startLoc - expectedLoc;
                let text2;
                if (endLoc === -1) {
                    text2 = text.substring(startLoc, startLoc + text1.length);
                }
                else {
                    text2 = text.substring(startLoc, endLoc + this.matchMaxBits);
                }
                if (text1 === text2) {
                    text = text.substring(0, startLoc)
                        + this.diff_text2(patches[x].diffs)
                        + text.substring(startLoc + text1.length);
                }
                else {
                    const diffs = this.diff_main(text1, text2, false);
                    if (text1.length > this.matchMaxBits &&
                        this.diff_levenshtein(diffs) / text1.length > this.patchDeleteThreshold) {
                        results[x] = false;
                    }
                    else {
                        this.diff_cleanupSemanticLossless(diffs);
                        let index1 = 0;
                        let index2 = 0;
                        for (let y = 0; y < patches[x].diffs.length; y++) {
                            const mod = patches[x].diffs[y];
                            if (mod[0] !== DiffOperation.DIFF_EQUAL) {
                                index2 = this.diff_xIndex(diffs, index1);
                            }
                            if (mod[0] === DiffOperation.DIFF_INSERT) {
                                text = text.substring(0, startLoc + index2) + mod[1]
                                    + text.substring(startLoc + index2);
                            }
                            else if (mod[0] === DiffOperation.DIFF_DELETE) {
                                text = text.substring(0, startLoc + index2)
                                    + text.substring(startLoc
                                        + this.diff_xIndex(diffs, index1 + mod[1].length));
                            }
                            if (mod[0] !== DiffOperation.DIFF_DELETE) {
                                index1 += mod[1].length;
                            }
                        }
                    }
                }
            }
        }
        text = text.substring(nullPadding.length, text.length - nullPadding.length);
        return [text, results];
    }
    patch_addPadding(patches) {
        const paddingLength = this.patchMargin;
        let nullPadding = "";
        for (let x = 1; x <= paddingLength; x++) {
            nullPadding += String.fromCharCode(x);
        }
        for (let x = 0; x < patches.length; x++) {
            patches[x].start1 += paddingLength;
            patches[x].start2 += paddingLength;
        }
        let patch = patches[0];
        let diffs = patch.diffs;
        if (diffs.length === 0 || diffs[0][0] !== DiffOperation.DIFF_EQUAL) {
            diffs.unshift([DiffOperation.DIFF_EQUAL, nullPadding]);
            patch.start1 -= paddingLength;
            patch.start2 -= paddingLength;
            patch.length1 += paddingLength;
            patch.length2 += paddingLength;
        }
        else if (paddingLength > diffs[0][1].length) {
            const extraLength = paddingLength - diffs[0][1].length;
            diffs[0][1] = nullPadding.substring(diffs[0][1].length) + diffs[0][1];
            patch.start1 -= extraLength;
            patch.start2 -= extraLength;
            patch.length1 += extraLength;
            patch.length2 += extraLength;
        }
        patch = patches[patches.length - 1];
        diffs = patch.diffs;
        if (diffs.length === 0 || diffs[diffs.length - 1][0] !== DiffOperation.DIFF_EQUAL) {
            diffs.push([DiffOperation.DIFF_EQUAL, nullPadding]);
            patch.length1 += paddingLength;
            patch.length2 += paddingLength;
        }
        else if (paddingLength > diffs[diffs.length - 1][1].length) {
            const extraLength = paddingLength - diffs[diffs.length - 1][1].length;
            diffs[diffs.length - 1][1] += nullPadding.substring(0, extraLength);
            patch.length1 += extraLength;
            patch.length2 += extraLength;
        }
        return nullPadding;
    }
    patch_splitMax(patches) {
        const patchSize = this.matchMaxBits;
        for (let x = 0; x < patches.length; x++) {
            if (patches[x].length1 <= patchSize) {
                continue;
            }
            const bigpatch = patches[x];
            patches.splice(x--, 1);
            let start1 = bigpatch.start1;
            let start2 = bigpatch.start2;
            let precontext = "";
            while (bigpatch.diffs.length !== 0) {
                const patch = new PatchObject();
                let empty = true;
                patch.start1 = start1 - precontext.length;
                patch.start2 = start2 - precontext.length;
                if (precontext !== "") {
                    patch.length1 = patch.length2 = precontext.length;
                    patch.diffs.push([DiffOperation.DIFF_EQUAL, precontext]);
                }
                while (bigpatch.diffs.length !== 0 &&
                    patch.length1 < patchSize - this.patchMargin) {
                    const diffType = bigpatch.diffs[0][0];
                    let diffText = bigpatch.diffs[0][1];
                    if (diffType === DiffOperation.DIFF_INSERT) {
                        patch.length2 += diffText.length;
                        start2 += diffText.length;
                        patch.diffs.push(bigpatch.diffs.shift());
                        empty = false;
                    }
                    else if (diffType === DiffOperation.DIFF_DELETE &&
                        patch.diffs.length === 1 &&
                        patch.diffs[0][0] === DiffOperation.DIFF_EQUAL &&
                        diffText.length > 2 * patchSize) {
                        patch.length1 += diffText.length;
                        start1 += diffText.length;
                        empty = false;
                        patch.diffs.push([diffType, diffText]);
                        bigpatch.diffs.shift();
                    }
                    else {
                        diffText = diffText.substring(0, patchSize - patch.length1 - this.patchMargin);
                        patch.length1 += diffText.length;
                        start1 += diffText.length;
                        if (diffType === DiffOperation.DIFF_EQUAL) {
                            patch.length2 += diffText.length;
                            start2 += diffText.length;
                        }
                        else {
                            empty = false;
                        }
                        patch.diffs.push([diffType, diffText]);
                        if (diffText === bigpatch.diffs[0][1]) {
                            bigpatch.diffs.shift();
                        }
                        else {
                            bigpatch.diffs[0][1] = bigpatch.diffs[0][1].substring(diffText.length);
                        }
                    }
                }
                precontext = this.diff_text2(patch.diffs);
                precontext = precontext.substring(precontext.length - this.patchMargin);
                const postcontext = this.diff_text1(bigpatch.diffs).substring(0, this.patchMargin);
                if (postcontext !== "") {
                    patch.length1 += postcontext.length;
                    patch.length2 += postcontext.length;
                    if (patch.diffs.length !== 0 &&
                        patch.diffs[patch.diffs.length - 1][0] === DiffOperation.DIFF_EQUAL) {
                        patch.diffs[patch.diffs.length - 1][1] += postcontext;
                    }
                    else {
                        patch.diffs.push([DiffOperation.DIFF_EQUAL, postcontext]);
                    }
                }
                if (!empty) {
                    patches.splice(++x, 0, patch);
                }
            }
        }
    }
    patch_toText(patches) {
        const text = [];
        for (let x = 0; x < patches.length; x++) {
            text[x] = patches[x];
        }
        return text.join("");
    }
    patch_fromText(textline) {
        const patches = [];
        if (!textline) {
            return patches;
        }
        const text = textline.split("\n");
        let textPointer = 0;
        const patchHeader = /^@@ -(\d+),?(\d*) \+(\d+),?(\d*) @@$/;
        while (textPointer < text.length) {
            const m = text[textPointer].match(patchHeader);
            if (!m) {
                throw new Error("Invalid patch string: " + text[textPointer]);
            }
            const patch = new PatchObject();
            patches.push(patch);
            patch.start1 = parseInt(m[1], 10);
            if (m[2] === "") {
                patch.start1--;
                patch.length1 = 1;
            }
            else if (m[2] === "0") {
                patch.length1 = 0;
            }
            else {
                patch.start1--;
                patch.length1 = parseInt(m[2], 10);
            }
            patch.start2 = parseInt(m[3], 10);
            if (m[4] === "") {
                patch.start2--;
                patch.length2 = 1;
            }
            else if (m[4] === "0") {
                patch.length2 = 0;
            }
            else {
                patch.start2--;
                patch.length2 = parseInt(m[4], 10);
            }
            textPointer++;
            let sign;
            let line;
            let rawLine;
            while (textPointer < text.length) {
                sign = text[textPointer].charAt(0);
                rawLine = text[textPointer].substring(1);
                try {
                    line = decodeURI(rawLine);
                }
                catch {
                    throw new Error("Illegal escape in patch_fromText: " + rawLine);
                }
                if (sign === "-") {
                    patch.diffs.push([DiffOperation.DIFF_DELETE, line]);
                }
                else if (sign === "+") {
                    patch.diffs.push([DiffOperation.DIFF_INSERT, line]);
                }
                else if (sign === " ") {
                    patch.diffs.push([DiffOperation.DIFF_EQUAL, line]);
                }
                else if (sign === "@") {
                    break;
                }
                else if (sign === "") {
                }
                else {
                    throw new Error('Invalid patch mode "' + sign + '" in: ' + line);
                }
                textPointer++;
            }
        }
        return patches;
    }
    diff_compute_(text1, text2, checklines, deadline) {
        let diffs;
        if (!text1) {
            return [[DiffOperation.DIFF_INSERT, text2]];
        }
        if (!text2) {
            return [[DiffOperation.DIFF_DELETE, text1]];
        }
        const longtext = text1.length > text2.length ? text1 : text2;
        const shorttext = text1.length > text2.length ? text2 : text1;
        const i = longtext.indexOf(shorttext);
        if (i !== -1) {
            diffs = [
                [DiffOperation.DIFF_INSERT, longtext.substring(0, i)],
                [DiffOperation.DIFF_EQUAL, shorttext],
                [DiffOperation.DIFF_INSERT, longtext.substring(i + shorttext.length)]
            ];
            if (text1.length > text2.length) {
                diffs[0][0] = DiffOperation.DIFF_DELETE;
                diffs[2][0] = DiffOperation.DIFF_DELETE;
            }
            return diffs;
        }
        if (shorttext.length === 1) {
            return [
                [DiffOperation.DIFF_DELETE, text1],
                [DiffOperation.DIFF_INSERT, text2]
            ];
        }
        const hm = this.diff_halfMatch_(text1, text2);
        if (hm) {
            const text1A = hm[0];
            const text1B = hm[1];
            const text2A = hm[2];
            const text2B = hm[3];
            const midCommon = hm[4];
            const diffsA = this.diff_main(text1A, text2A, checklines, deadline);
            const diffsB = this.diff_main(text1B, text2B, checklines, deadline);
            return diffsA.concat([[DiffOperation.DIFF_EQUAL, midCommon]], diffsB);
        }
        if (checklines && text1.length > 100 && text2.length > 100) {
            return this.diff_lineMode_(text1, text2, deadline);
        }
        return this.diff_bisect_(text1, text2, deadline);
    }
    diff_lineMode_(text1, text2, deadline) {
        const a = this.diff_linesToChars(text1, text2);
        text1 = a.chars1;
        text2 = a.chars2;
        const linearray = a.lineArray;
        const diffs = this.diff_main(text1, text2, false, deadline);
        this.diff_charsToLines(diffs, linearray);
        this.diff_cleanupSemantic(diffs);
        diffs.push([DiffOperation.DIFF_EQUAL, ""]);
        let pointer = 0;
        let countDelete = 0;
        let countInsert = 0;
        let textDelete = "";
        let textInsert = "";
        while (pointer < diffs.length) {
            switch (diffs[pointer][0]) {
                case DiffOperation.DIFF_INSERT:
                    countInsert++;
                    textInsert += diffs[pointer][1];
                    break;
                case DiffOperation.DIFF_DELETE:
                    countDelete++;
                    textDelete += diffs[pointer][1];
                    break;
                case DiffOperation.DIFF_EQUAL:
                    if (countDelete >= 1 && countInsert >= 1) {
                        diffs.splice(pointer - countDelete - countInsert, countDelete + countInsert);
                        pointer = pointer - countDelete - countInsert;
                        const subDiff = this.diff_main(textDelete, textInsert, false, deadline);
                        for (let j = subDiff.length - 1; j >= 0; j--) {
                            diffs.splice(pointer, 0, subDiff[j]);
                        }
                        pointer = pointer + subDiff.length;
                    }
                    countInsert = 0;
                    countDelete = 0;
                    textDelete = "";
                    textInsert = "";
                    break;
            }
            pointer++;
        }
        diffs.pop();
        return diffs;
    }
    diff_bisect_(text1, text2, deadline) {
        const text1Length = text1.length;
        const text2Length = text2.length;
        const maxD = Math.ceil((text1Length + text2Length) / 2);
        const vOffset = maxD;
        const vLength = 2 * maxD;
        const v1 = new Array(vLength);
        const v2 = new Array(vLength);
        for (let x = 0; x < vLength; x++) {
            v1[x] = -1;
            v2[x] = -1;
        }
        v1[vOffset + 1] = 0;
        v2[vOffset + 1] = 0;
        const delta = text1Length - text2Length;
        const front = (delta % 2 !== 0);
        let k1Start = 0;
        let k1End = 0;
        let k2Start = 0;
        let k2End = 0;
        for (let d = 0; d < maxD; d++) {
            if (Date.now() > deadline) {
                break;
            }
            for (let k1 = -d + k1Start; k1 <= d - k1End; k1 += 2) {
                const k1Offset = vOffset + k1;
                let x1;
                if (k1 === -d || (k1 !== d && v1[k1Offset - 1] < v1[k1Offset + 1])) {
                    x1 = v1[k1Offset + 1];
                }
                else {
                    x1 = v1[k1Offset - 1] + 1;
                }
                let y1 = x1 - k1;
                while (x1 < text1Length
                    && y1 < text2Length
                    && text1.charAt(x1) === text2.charAt(y1)) {
                    x1++;
                    y1++;
                }
                v1[k1Offset] = x1;
                if (x1 > text1Length) {
                    k1End += 2;
                }
                else if (y1 > text2Length) {
                    k1Start += 2;
                }
                else if (front) {
                    const k2Offset = vOffset + delta - k1;
                    if (k2Offset >= 0 && k2Offset < vLength && v2[k2Offset] !== -1) {
                        const x2 = text1Length - v2[k2Offset];
                        if (x1 >= x2) {
                            return this.diff_bisectSplit_(text1, text2, x1, y1, deadline);
                        }
                    }
                }
            }
            for (let k2 = -d + k2Start; k2 <= d - k2End; k2 += 2) {
                const k2Offset = vOffset + k2;
                let x2;
                if (k2 === -d || (k2 !== d && v2[k2Offset - 1] < v2[k2Offset + 1])) {
                    x2 = v2[k2Offset + 1];
                }
                else {
                    x2 = v2[k2Offset - 1] + 1;
                }
                let y2 = x2 - k2;
                while (x2 < text1Length
                    && y2 < text2Length
                    && text1.charAt(text1Length - x2 - 1) === text2.charAt(text2Length - y2 - 1)) {
                    x2++;
                    y2++;
                }
                v2[k2Offset] = x2;
                if (x2 > text1Length) {
                    k2End += 2;
                }
                else if (y2 > text2Length) {
                    k2Start += 2;
                }
                else if (!front) {
                    const k1Offset = vOffset + delta - k2;
                    if (k1Offset >= 0 && k1Offset < vLength && v1[k1Offset] !== -1) {
                        const x1 = v1[k1Offset];
                        const y1 = vOffset + x1 - k1Offset;
                        x2 = text1Length - x2;
                        if (x1 >= x2) {
                            return this.diff_bisectSplit_(text1, text2, x1, y1, deadline);
                        }
                    }
                }
            }
        }
        return [
            [DiffOperation.DIFF_DELETE, text1],
            [DiffOperation.DIFF_INSERT, text2]
        ];
    }
    diff_bisectSplit_(text1, text2, x, y, deadline) {
        const text1A = text1.substring(0, x);
        const text2A = text2.substring(0, y);
        const text1B = text1.substring(x);
        const text2B = text2.substring(y);
        const diffsA = this.diff_main(text1A, text2A, false, deadline);
        const diffsB = this.diff_main(text1B, text2B, false, deadline);
        return diffsA.concat(diffsB);
    }
    diff_linesToCharsMunge_(text, lineArray, lineHash, maxLines) {
        let chars = "";
        let lineStart = 0;
        let lineEnd = -1;
        let lineArrayLength = lineArray.length;
        while (lineEnd < text.length - 1) {
            lineEnd = text.indexOf("\n", lineStart);
            if (lineEnd === -1) {
                lineEnd = text.length - 1;
            }
            let line = text.substring(lineStart, lineEnd + 1);
            if (lineHash.hasOwnProperty
                ? lineHash.hasOwnProperty(line)
                : (lineHash[line] !== undefined)) {
                chars += String.fromCharCode(lineHash[line]);
            }
            else {
                if (lineArrayLength === maxLines) {
                    line = text.substring(lineStart);
                    lineEnd = text.length;
                }
                chars += String.fromCharCode(lineArrayLength);
                lineHash[line] = lineArrayLength;
                lineArray[lineArrayLength++] = line;
            }
            lineStart = lineEnd + 1;
        }
        return chars;
    }
    diff_commonOverlap_(text1, text2) {
        const text1Length = text1.length;
        const text2Length = text2.length;
        if (text1Length === 0 || text2Length === 0) {
            return 0;
        }
        if (text1Length > text2Length) {
            text1 = text1.substring(text1Length - text2Length);
        }
        else if (text1Length < text2Length) {
            text2 = text2.substring(0, text1Length);
        }
        const textLength = math.min(text1Length, text2Length);
        if (text1 === text2) {
            return textLength;
        }
        let best = 0;
        let length = 1;
        while (true) {
            const pattern = text1.substring(textLength - length);
            const found = text2.indexOf(pattern);
            if (found === -1) {
                return best;
            }
            length += found;
            if (found === 0 ||
                text1.substring(textLength - length) === text2.substring(0, length)) {
                best = length;
                length++;
            }
        }
    }
    diff_halfMatch_(text1, text2) {
        if (this.diffTimeout <= 0) {
            return null;
        }
        const longtext = text1.length > text2.length ? text1 : text2;
        const shorttext = text1.length > text2.length ? text2 : text1;
        if (longtext.length < 4 || shorttext.length * 2 < longtext.length) {
            return null;
        }
        const hm1 = this.diff_halfMatchI_(longtext, shorttext, Math.ceil(longtext.length / 4));
        const hm2 = this.diff_halfMatchI_(longtext, shorttext, Math.ceil(longtext.length / 2));
        let hm;
        if (!hm1 && !hm2) {
            return null;
        }
        else if (!hm2) {
            hm = hm1;
        }
        else if (!hm1) {
            hm = hm2;
        }
        else {
            hm = hm1[4].length > hm2[4].length ? hm1 : hm2;
        }
        let text1A;
        let text1B;
        let text2A;
        let text2B;
        const midCommon = hm[4];
        if (text1.length > text2.length) {
            text1A = hm[0];
            text1B = hm[1];
            text2A = hm[2];
            text2B = hm[3];
        }
        else {
            text2A = hm[0];
            text2B = hm[1];
            text1A = hm[2];
            text1B = hm[3];
        }
        return [text1A, text1B, text2A, text2B, midCommon];
    }
    diff_halfMatchI_(longtext, shorttext, i) {
        const seed = longtext.substring(i, i + Math.floor(longtext.length / 4));
        let bestCommon = "";
        let bestLongtextA;
        let bestLongtextB;
        let bestShorttextA;
        let bestShorttextB;
        let j = shorttext.indexOf(seed, 0);
        while (j !== -1) {
            const prefixLength = this.diff_commonPrefix(longtext.substring(i), shorttext.substring(j));
            const suffixLength = this.diff_commonSuffix(longtext.substring(0, i), shorttext.substring(0, j));
            if (bestCommon.length < suffixLength + prefixLength) {
                bestCommon = shorttext.substring(j - suffixLength, j)
                    + shorttext.substring(j, j + prefixLength);
                bestLongtextA = longtext.substring(0, i - suffixLength);
                bestLongtextB = longtext.substring(i + prefixLength);
                bestShorttextA = shorttext.substring(0, j - suffixLength);
                bestShorttextB = shorttext.substring(j + prefixLength);
            }
            j = shorttext.indexOf(seed, j + 1);
        }
        if (bestCommon.length * 2 >= longtext.length) {
            return [
                bestLongtextA,
                bestLongtextB,
                bestShorttextA,
                bestShorttextB,
                bestCommon
            ];
        }
        return null;
    }
    diff_cleanupSemanticScore_(one, two) {
        if (!one || !two) {
            return 6;
        }
        const char1 = one.charAt(one.length - 1);
        const char2 = two.charAt(0);
        const nonAlphaNumeric1 = char1.match(NON_ALPHA_NUMERIC_REGEX);
        const nonAlphaNumeric2 = char2.match(NON_ALPHA_NUMERIC_REGEX);
        const whitespace1 = nonAlphaNumeric1 && char1.match(WHITESPACE_REGEX);
        const whitespace2 = nonAlphaNumeric2 && char2.match(WHITESPACE_REGEX);
        const lineBreak1 = whitespace1 && char1.match(LINEBREAK_REGEX);
        const lineBreak2 = whitespace2 && char2.match(LINEBREAK_REGEX);
        const blankLine1 = lineBreak1 && one.match(BLANKLINE_END_REGEX);
        const blankLine2 = lineBreak2 && two.match(BLANKLINE_START_REGEX);
        if (blankLine1 || blankLine2) {
            return 5;
        }
        else if (lineBreak1 || lineBreak2) {
            return 4;
        }
        else if (nonAlphaNumeric1 && !whitespace1 && whitespace2) {
            return 3;
        }
        else if (whitespace1 || whitespace2) {
            return 2;
        }
        else if (nonAlphaNumeric1 || nonAlphaNumeric2) {
            return 1;
        }
        return 0;
    }
    match_bitap_(text, pattern, loc) {
        if (pattern.length > this.matchMaxBits) {
            throw new Error("Pattern too long for this browser");
        }
        const s = this.match_alphabet_(pattern);
        let scoreThreshold = this.matchThreshold;
        let bestLoc = text.indexOf(pattern, loc);
        if (bestLoc !== -1) {
            scoreThreshold = math.min(this.match_bitapScore_(0, bestLoc, pattern, loc), scoreThreshold);
            bestLoc = text.lastIndexOf(pattern, loc + pattern.length);
            if (bestLoc !== -1) {
                scoreThreshold = math.min(this.match_bitapScore_(0, bestLoc, pattern, loc), scoreThreshold);
            }
        }
        const matchmask = 1 << (pattern.length - 1);
        bestLoc = -1;
        let binMin;
        let binMid;
        let binMax = pattern.length + text.length;
        let lastRD;
        for (let d = 0; d < pattern.length; d++) {
            binMin = 0;
            binMid = binMax;
            while (binMin < binMid) {
                if (this.match_bitapScore_(d, loc + binMid, pattern, loc) <= scoreThreshold) {
                    binMin = binMid;
                }
                else {
                    binMax = binMid;
                }
                binMid = Math.floor((binMax - binMin) / 2 + binMin);
            }
            binMax = binMid;
            let start = math.max(1, loc - binMid + 1);
            const finish = math.min(loc + binMid, text.length) + pattern.length;
            const rd = Array(finish + 2);
            rd[finish + 1] = (1 << d) - 1;
            for (let j = finish; j >= start; j--) {
                const charMatch = s[text.charAt(j - 1)];
                if (d === 0) {
                    rd[j] = ((rd[j + 1] << 1) | 1) & charMatch;
                }
                else {
                    rd[j] = (((rd[j + 1] << 1) | 1) & charMatch) |
                        (((lastRD[j + 1] | lastRD[j]) << 1) | 1) |
                        lastRD[j + 1];
                }
                if (rd[j] & matchmask) {
                    const score = this.match_bitapScore_(d, j - 1, pattern, loc);
                    if (score <= scoreThreshold) {
                        scoreThreshold = score;
                        bestLoc = j - 1;
                        if (bestLoc > loc) {
                            start = math.max(1, 2 * loc - bestLoc);
                        }
                        else {
                            break;
                        }
                    }
                }
            }
            if (this.match_bitapScore_(d + 1, loc, pattern, loc) > scoreThreshold) {
                break;
            }
            lastRD = rd;
        }
        return bestLoc;
    }
    match_bitapScore_(e, x, pattern, loc) {
        const accuracy = e / pattern.length;
        const proximity = Math.abs(loc - x);
        if (!this.matchDistance) {
            return proximity ? 1.0 : accuracy;
        }
        return accuracy + (proximity / this.matchDistance);
    }
    match_alphabet_(pattern) {
        const s = {};
        for (let i = 0; i < pattern.length; i++) {
            s[pattern.charAt(i)] = 0;
        }
        for (let i = 0; i < pattern.length; i++) {
            s[pattern.charAt(i)] |= 1 << (pattern.length - i - 1);
        }
        return s;
    }
    patch_addContext_(patch, text) {
        if (text.length === 0) {
            return;
        }
        if (patch.start2 == null) {
            throw Error("patch not initialized");
        }
        let pattern = text.substring(patch.start2, patch.start2 + patch.length1);
        let padding = 0;
        while (text.indexOf(pattern) !== text.lastIndexOf(pattern) &&
            pattern.length < (this.matchMaxBits - this.patchMargin - this.patchMargin)) {
            padding += this.patchMargin;
            pattern = text.substring(patch.start2 - padding, patch.start2 + patch.length1 + padding);
        }
        padding += this.patchMargin;
        const prefix = text.substring(patch.start2 - padding, patch.start2);
        if (prefix) {
            patch.diffs.unshift([DiffOperation.DIFF_EQUAL, prefix]);
        }
        const suffix = text.substring(patch.start2 + patch.length1, patch.start2 + patch.length1 + padding);
        if (suffix) {
            patch.diffs.push([DiffOperation.DIFF_EQUAL, suffix]);
        }
        patch.start1 -= prefix.length;
        patch.start2 -= prefix.length;
        patch.length1 += prefix.length + suffix.length;
        patch.length2 += prefix.length + suffix.length;
    }
}
