import { PatchObject } from "./PatchObject";
import type { Diff, PatchApplyArray } from "../types";
/**
 * Diff Match and Patch
 * Copyright 2018 The diff-match-patch Authors.
 *
 * https://github.com/google/diff-match-patch
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @fileoverview Computes the difference between two texts to create a patch.
 * Applies the patch onto another text, allowing for errors.
 * @author fraser@google.com (Neil Fraser)
 *
 * Ported by [xiaochao.k@gmail.com](https://github.com/nonoroazoro)
 */
/**
 * Class containing the diff, match and patch methods.
 */
export declare class DiffMatchPatch {
    /**
     * Number of seconds to map a diff before giving up (0 for infinity).
     */
    diffTimeout: number;
    /**
     * Cost of an empty edit operation in terms of edit characters.
     */
    diffEditCost: number;
    /**
     * At what point is no match declared (0.0 = perfection, 1.0 = very loose).
     */
    matchThreshold: number;
    /**
     * How far to search for a match (0 = exact location, 1000+ = broad match).
     * A match this many characters away from the expected location will add
     * 1.0 to the score (0.0 is a perfect match).
     */
    matchDistance: number;
    /**
     * When deleting a large block of text (over ~64 characters), how close do
     * the contents have to be to match the expected contents. (0.0 = perfection,
     * 1.0 = very loose). Note that Match_Threshold controls how closely the
     * end points of a delete need to match.
     */
    patchDeleteThreshold: number;
    /**
     * Chunk size for context length.
     */
    patchMargin: number;
    /**
     * The number of bits in an int.
     */
    matchMaxBits: number;
    /**
     * Find the differences between two texts. Simplifies the problem by stripping
     * any common prefix or suffix off the texts before diffing.
     *
     * @param {string} text1 Old string to be diffed.
     * @param {string} text2 New string to be diffed.
     * @param {boolean} [optChecklines] Optional speedup flag. If present and false,
     * then don't run a line-level diff first to identify the changed areas.
     * Defaults to true, which does a faster, slightly less optimal diff.
     * @param {number} [optDeadline] Optional time when the diff should be complete
     * by. Used internally for recursive calls. Users should set DiffTimeout
     * instead.
     * @returns {Diff[]} Array of diff tuples.
     */
    diff_main(text1: string, text2: string, optChecklines?: boolean, optDeadline?: number): Diff[];
    /**
     * Determine the common prefix of two strings.
     *
     * @param {string} text1 First string.
     * @param {string} text2 Second string.
     * @returns {number} The number of characters common to the start of each string.
     */
    diff_commonPrefix(text1: string, text2: string): number;
    /**
     * Determine the common suffix of two strings.
     *
     * @param {string} text1 First string.
     * @param {string} text2 Second string.
     * @returns {number} The number of characters common to the end of each string.
     */
    diff_commonSuffix(text1: string, text2: string): number;
    /**
     * Reduce the number of edits by eliminating semantically trivial equalities.
     *
     * @param {Diff[]} diffs Array of diff tuples.
     */
    diff_cleanupSemantic(diffs: Diff[]): void;
    /**
     * Look for single edits surrounded on both sides by equalities
     * which can be shifted sideways to align the edit to a word boundary.
     * e.g: The c<ins>at c</ins>ame. -> The <ins>cat </ins>came.
     *
     * @param {Diff[]} diffs Array of diff tuples.
     */
    diff_cleanupSemanticLossless(diffs: Diff[]): void;
    /**
     * Reduce the number of edits by eliminating operationally trivial equalities.
     *
     * @param {Diff[]} diffs Array of diff tuples.
     */
    diff_cleanupEfficiency(diffs: Diff[]): void;
    /**
     * Reorder and merge like edit sections. Merge equalities.
     * Any edit section can move as long as it doesn't cross an equality.
     *
     * @param {Diff[]} diffs Array of diff tuples.
     */
    diff_cleanupMerge(diffs: Diff[]): void;
    /**
     * loc is a location in text1, compute and return the equivalent location in
     * text2.
     * e.g. 'The cat' vs 'The big cat', 1->1, 5->8
     *
     * @param {Diff[]} diffs Array of diff tuples.
     * @param {number} loc Location within text1.
     * @returns {number} Location within text2.
     */
    diff_xIndex(diffs: Diff[], loc: number): number;
    /**
     * Convert a diff array into a pretty HTML report.
     *
     * @param {Diff[]>} diffs Array of diff tuples.
     * @returns {string} HTML representation.
     */
    diff_prettyHtml(diffs: Diff[]): string;
    /**
     * Compute and return the source text (all equalities and deletions).
     *
     * @param {Diff[]} diffs Array of diff tuples.
     * @returns {string} Source text.
     */
    diff_text1(diffs: Diff[]): string;
    /**
     * Compute and return the destination text (all equalities and insertions).
     *
     * @param {Diff[]} diffs Array of diff tuples.
     * @returns {string} Destination text.
     */
    diff_text2(diffs: Diff[]): string;
    /**
     * Compute the Levenshtein distance; the number of inserted, deleted or
     * substituted characters.
     *
     * @param {Diff[]} diffs Array of diff tuples.
     * @returns {number} Number of changes.
     */
    diff_levenshtein(diffs: Diff[]): number;
    /**
     * Crush the diff into an encoded string which describes the operations
     * required to transform text1 into text2.
     * E.g. =3\t-2\t+ing  -> Keep 3 chars, delete 2 chars, insert 'ing'.
     * Operations are tab-separated. Inserted text is escaped using %xx notation.
     *
     * @param {Diff[]} diffs Array of diff tuples.
     * @returns {string} Delta text.
     */
    diff_toDelta(diffs: Diff[]): string;
    /**
     * Given the original text1, and an encoded string which describes the
     * operations required to transform text1 into text2, compute the full diff.
     *
     * @param {string} text1 Source string for the diff.
     * @param {string} delta Delta text.
     * @returns {Diff[]} Array of diff tuples.
     * @throws {Error} If invalid input.
     */
    diff_fromDelta(text1: string, delta: string): Diff[];
    /**
     * Split two texts into an array of strings. Reduce the texts to a string of
     * hashes where each Unicode character represents one line.
     *
     * @param {string} text1 First string.
     * @param {string} text2 Second string.
     * @returns {{chars1: string, chars2: string, lineArray: string[]}}
     * An object containing the encoded text1, the encoded text2 and
     * the array of unique strings.
     * The zeroth element of the array of unique strings is intentionally blank.
     */
    diff_linesToChars(text1: string, text2: string): {
        chars1: string;
        chars2: string;
        lineArray: string[];
    };
    /**
     * Rehydrate the text in a diff from a string of line hashes to real lines of
     * text.
     *
     * @param {Diff[]} diffs Array of diff tuples.
     * @param {string[]} lineArray Array of unique strings.
     */
    diff_charsToLines(diffs: Diff[], lineArray: string[]): void;
    /**
     * Locate the best instance of 'pattern' in 'text' near 'loc'.
     *
     * @param {string} text The text to search.
     * @param {string} pattern The pattern to search for.
     * @param {number} loc The location to search around.
     * @returns {number} Best match index or -1.
     */
    match_main(text: string, pattern: string, loc: number): number;
    /**
     * Compute a list of patches to turn text1 into text2.
     * Use diffs if provided, otherwise compute it ourselves.
     * There are four ways to call this function, depending on what data is
     * available to the caller:
     * Method 1:
     * a = text1, b = text2
     * Method 2:
     * a = diffs
     * Method 3 (optimal):
     * a = text1, b = diffs
     * Method 4 (deprecated, use method 3):
     * a = text1, b = text2, c = diffs
     *
     * @param {(string|Diff[])} a text1 (methods 1,3,4) or
     * Array of diff tuples for text1 to text2 (method 2).
     * @param {(string|Diff[])} [optB] text2 (methods 1,4) or
     * Array of diff tuples for text1 to text2 (method 3) or undefined (method 2).
     * @param {(string|Diff[])} [optC] Array of diff tuples
     * for text1 to text2 (method 4) or undefined (methods 1,2,3).
     * @returns {PatchObject[]} Array of Patch objects.
     */
    patch_make(a: string | Diff[], optB?: string | Diff[], optC?: string | Diff[]): PatchObject[];
    /**
     * Given an array of patches, return another array that is identical.
     *
     * @param {PatchObject[]>} patches Array of Patch objects.
     * @returns {PatchObject[]} Array of Patch objects.
     */
    patch_deepCopy(patches: PatchObject[]): PatchObject[];
    /**
     * Merge a set of patches onto the text. Return a patched text, as well
     * as a list of true/false values indicating which patches were applied.
     *
     * @param {PatchObject[]} patches Array of Patch objects.
     * @param {string} text Old text.
     * @returns {PatchApplyArray} Two element Array, containing the
     * new text and an array of boolean values.
     */
    patch_apply(patches: PatchObject[], text: string): PatchApplyArray;
    /**
     * Add some padding on text start and end so that edges can match something.
     * Intended to be called only from within patch_apply.
     *
     * @param {PatchObject[]} patches Array of Patch objects.
     * @returns {string} The padding string added to each side.
     */
    patch_addPadding(patches: PatchObject[]): string;
    /**
     * Look through the patches and break up any which are longer than the maximum
     * limit of the match algorithm.
     * Intended to be called only from within patch_apply.
     *
     * @param {PatchObject[]} patches Array of Patch objects.
     */
    patch_splitMax(patches: PatchObject[]): void;
    /**
     * Take a list of patches and return a textual representation.
     *
     * @param {PatchObject[]} patches Array of Patch objects.
     * @returns {string} Text representation of patches.
     */
    patch_toText(patches: PatchObject[]): string;
    /**
     * Parse a textual representation of patches and return a list of Patch objects.
     *
     * @param {string} textline Text representation of patches.
     * @returns {PatchObject[]} Array of Patch objects.
     * @throws {Error} If invalid input.
     */
    patch_fromText(textline: string): PatchObject[];
    /**
     * Find the differences between two texts. Assumes that the texts do not
     * have any common prefix or suffix.
     *
     * @private
     * @param {string} text1 Old string to be diffed.
     * @param {string} text2 New string to be diffed.
     * @param {boolean} checklines Speedup flag. If false, then don't run a
     * line-level diff first to identify the changed areas.
     * If true, then run a faster, slightly less optimal diff.
     * @param {number} deadline Time when the diff should be complete by.
     * @returns {Diff[]} Array of diff tuples.
     */
    private diff_compute_;
    /**
     * Do a quick line-level diff on both strings, then re-diff the parts for
     * greater accuracy.
     * This speedup can produce non-minimal diffs.
     *
     * @private
     * @param {string} text1 Old string to be diffed.
     * @param {string} text2 New string to be diffed.
     * @param {number} deadline Time when the diff should be complete by.
     * @returns {Diff[]} Array of diff tuples.
     */
    private diff_lineMode_;
    /**
     * Find the 'middle snake' of a diff, split the problem in two
     * and return the recursively constructed diff.
     * See Myers 1986 paper: An O(ND) Difference Algorithm and Its Variations.
     *
     * @private
     * @param {string} text1 Old string to be diffed.
     * @param {string} text2 New string to be diffed.
     * @param {number} deadline Time at which to bail if not yet complete.
     * @returns {Diff[]} Array of diff tuples.
     */
    private diff_bisect_;
    /**
     * Given the location of the 'middle snake', split the diff in two parts
     * and recurse.
     *
     * @private
     * @param {string} text1 Old string to be diffed.
     * @param {string} text2 New string to be diffed.
     * @param {number} x Index of split point in text1.
     * @param {number} y Index of split point in text2.
     * @param {number} deadline Time at which to bail if not yet complete.
     * @returns {Diff[]} Array of diff tuples.
     */
    private diff_bisectSplit_;
    /**
     * Split a text into an array of strings. Reduce the texts to a string of
     * hashes where each Unicode character represents one line.
     * Modifies linearray and linehash through being a closure.
     *
     * @private
     * @param {string} text String to encode.
     * @param {string[]} lineArray Array of unique strings.
     * @param {Record<string, number>} lineHash Line-hash pairs.
     * @param {number} maxLines
     * @returns {string} Encoded string.
     */
    private diff_linesToCharsMunge_;
    /**
     * Determine if the suffix of one string is the prefix of another.
     *
     * @private
     * @param {string} text1 First string.
     * @param {string} text2 Second string.
     * @returns {number} The number of characters common to the end of the first
     * string and the start of the second string.
     */
    private diff_commonOverlap_;
    /**
     * Do the two texts share a substring which is at least half the length of the
     * longer text?
     * This speedup can produce non-minimal diffs.
     *
     * @private
     * @param {string} text1 First string.
     * @param {string} text2 Second string.
     * @returns {(HalfMatchArray | null)} Five element Array, containing the prefix of
     * text1, the suffix of text1, the prefix of text2, the suffix of
     * text2 and the common middle. Or null if there was no match.
     */
    private diff_halfMatch_;
    /**
     * Does a substring of shorttext exist within longtext such that the substring
     * is at least half the length of longtext?
     * Closure, but does not reference any external variables.
     *
     * @private
     * @param {string} longtext Longer string.
     * @param {string} shorttext Shorter string.
     * @param {number} i Start index of quarter length substring within longtext.
     * @returns {(HalfMatchArray | null)} Five element Array, containing the prefix of
     * longtext, the suffix of longtext, the prefix of shorttext, the suffix
     * of shorttext and the common middle. Or null if there was no match.
     */
    private diff_halfMatchI_;
    /**
     * Given two strings, compute a score representing whether the internal
     * boundary falls on logical boundaries.
     * Scores range from 6 (best) to 0 (worst).
     * Closure, but does not reference any external variables.
     *
     * @private
     * @param {string} one First string.
     * @param {string} two Second string.
     * @returns {number} The score.
     */
    private diff_cleanupSemanticScore_;
    /**
     * Locate the best instance of 'pattern' in 'text' near 'loc' using the
     * Bitap algorithm.
     *
     * @private
     * @param {string} text The text to search.
     * @param {string} pattern The pattern to search for.
     * @param {number} loc The location to search around.
     * @returns {number} Best match index or -1.
     */
    private match_bitap_;
    /**
     * Compute and return the score for a match with e errors and x location.
     * Accesses loc and pattern through being a closure.
     *
     * @private
     * @param {number} e Number of errors in match.
     * @param {number} x Location of match.
     * @param {string} pattern The pattern to search for.
     * @param {number} loc The location to search around.
     * @returns {number} Overall score for match (0.0 = good, 1.0 = bad).
     */
    private match_bitapScore_;
    /**
     * Initialize the alphabet for the Bitap algorithm.
     *
     * @private
     * @param {string} pattern The text to encode.
     * @returns {Record<string, number>} Hash of character locations.
     */
    private match_alphabet_;
    /**
     * Increase the context until it is unique,
     * but don't let the pattern expand beyond Match_MaxBits.
     *
     * @private
     * @param {PatchObject} patch The patch to grow.
     * @param {string} text Source text.
     */
    private patch_addContext_;
}
